from __future__ import print_function
import logging
import time
import feedparser

try:
    # Python 2
    from urllib import urlencode
    from urllib import urlretrieve
except ImportError:
    # Python 3
    from urllib.parse import urlencode
    from urllib.request import urlretrieve

logger = logging.getLogger(__name__)


class Search(object):
    """
    Class to search and download abstracts from the arXiv

    Args:
        query (string):
        id_list (list): List of arXiv IDs to be downloaded
        max_results (int): The maximum number of abstracts that should be downloaded. Defaults to
            infinity, i.e., no limit at all
        sort_by (string): The arXiv field by which the result should be sorted
        sort_order (string): The sorting order, i.e. "ascending", "descending" or None.
        max_results_per_call (int): Internally, a arXiv search query is split up into smaller
            queries that download the data iteratively in chunks. This parameter sets an upper bound
            on the number of abstracts to be downloaded in a single call from the arXiv API.
        time_sleep (int): Time (in seconds) between two subsequent arXiv REST calls. Defaults to
            :code:`3`, the recommendation of arXiv.
        prune (bool): Whether some of the keys in a downloaded arXiv result should be dropped.
            Defaults to True.

    """

    root_url = 'http://export.arxiv.org/api/'
    prune_keys = [
        'updated_parsed',
        'published_parsed',
        'arxiv_primary_category',
        'summary_detail',
        'author',
        'author_detail',
        'links',
        'guidislink',
        'title_detail',
        'tags',
        'id'
    ]

    def __init__(self, query=None, id_list=None, max_results=None, sort_by=None,
                 sort_order=None, max_results_per_call=None, time_sleep=3, prune=True):

        self.query = query
        self.id_list = id_list
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.max_results_per_call = max_results_per_call
        self.time_sleep = time_sleep
        self.prune = prune
        self.max_results = max_results

        if not self.max_results:
            logger.info('No maximal number of results given by the user. Download all')
            self.max_results = float('inf')

    def _get_url(self, start=0, max_results=None):

        url_args = urlencode(
            {
                "search_query": self.query,
                "id_list": self.id_list,
                "start": start,
                "max_results": max_results,
                "sortBy": self.sort_by,
                "sortOrder": self.sort_order
            }
        )

        return self.root_url + 'query?' + url_args

    def _parse(self, url):
        """
        Downloads the data provided by the REST endpoint given in the url.
        """
        result = feedparser.parse(url)

        if result.get('status') != 200:
            # TODO: better error reporting
            raise Exception(
                "HTTP Error {} in query".format(result.get('status', 'no status')))
        else:
            return result['entries']

    def _prune_result(self, result):
        """
        Deletes some of the keys from the downloaded result
        """

        for key in self.prune_keys:
            try:
                del result['key']
            except KeyError:
                pass

        return result

    def _process_result(self, result):

        # Useful to have for download automation
        result['pdf_url'] = None
        for link in result['links']:
            if 'title' in link and link['title'] == 'pdf':
                result['pdf_url'] = link['href']
        result['affiliation'] = result.pop('arxiv_affiliation', 'None')
        result['arxiv_url'] = result.pop('link')
        result['title'] = result['title'].rstrip('\n')
        result['summary'] = result['summary'].rstrip('\n')
        result['authors'] = [d['name'] for d in result['authors']]
        if 'arxiv_comment' in result:
            result['arxiv_comment'] = result['arxiv_comment'].rstrip('\n')
        else:
            result['arxiv_comment'] = None
        if 'arxiv_journal_ref' in result:
            result['journal_reference'] = result.pop('arxiv_journal_ref')
        else:
            result['journal_reference'] = None
        if 'arxiv_doi' in result:
            result['doi'] = result.pop('arxiv_doi')
        else:
            result['doi'] = None

        if self.prune:
            result = self._prune_result(result)

        return result

    def _get_next(self):

        n_left = self.max_results
        start = 0

        while n_left > 0:

            if n_left < self.max_results:
                logger.info('... play nice on the arXiv and sleep a bit ...')
                time.sleep(self.time_sleep)

            logger.info('Fetch from arxiv ({} results left to download)'.format(n_left))
            url = self._get_url(
                start=start,
                max_results=min(n_left, self.max_results_per_call))

            results = self._parse(url)

            # Update the entries left to download
            n_fetched = len(results)
            logger.info('Received {} entries'.format(n_fetched))

            if n_fetched == 0:
                logger.info('No more entries left to fetch')
                logger.info('Fetching finished.')
                break

            # Update the number of results left to download
            n_left = n_left - n_fetched
            start = start + n_fetched

            # Process results
            results = [self._process_result(r) for r in results]

            yield results

    def download(self, iterative=False):
        """
        Triggers the download of the result of the given search query.

        Args:
            iterative (bool): If true, then an iterator is returned, which allows to download the
                data iteratively. Otherwise, all the data is fetched first and then returned.

        Returns:
            iterable: Either a list or a general iterator holding the result of the search query.
        """
        logger.info('Start downloading')
        if iterative:

            logger.info('Build iterator')

            def iterator():
                logger.info('Start iterating')
                for result in self._get_next():
                    for entry in result:
                        yield entry
            return iterator
        else:
            results = list()
            for result in self._get_next():
                results = results + result
            return results


def query(search_query="", id_list=[], prune=True, max_results=None, sort_by="relevance",
          sort_order="descending", max_results_per_call=1000, iterative=False):
    """
    See :py:class:`arxiv.Search` for a description of the parameters.
    """

    search = Search(
        query=search_query,
        id_list=','.join(id_list),
        sort_by=sort_by,
        sort_order=sort_order,
        prune=prune,
        max_results=max_results,
        max_results_per_call=max_results_per_call)

    return search.download(iterative=iterative)


def to_slug(title):
    # Remove special characters
    filename = ''.join(c if c.isalnum() else '_' for c in title)
    # delete duplicate underscores
    filename = '_'.join(list(filter(None, filename.split('_'))))
    return filename


def download(obj, dirname='./', prepend_id=False, slugify=False):
    # Downloads file in obj (can be result or unique page) if it has a .pdf link
    if 'pdf_url' in obj and 'title' in obj and obj['pdf_url'] and obj['title']:
        filename = obj['title']
        if slugify:
            filename = to_slug(filename)
        if prepend_id:
            filename = obj['arxiv_url'].split('/')[-1] + '-' + filename
        filename = dirname + filename + '.pdf'
        # Download
        urlretrieve(obj['pdf_url'], filename)
        return filename
    else:
        print("Object obj has no PDF URL, or has no title")
